'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var path = require('path');
var tComm = require('t-comm');
var shared = require('@plugin-light/shared');
var fs = require('fs');
var webpackPluginDispatchVue = require('@plugin-light/webpack-plugin-dispatch-vue');
var sharedVue2 = require('@plugin-light/shared-vue2');
var webpackPluginAddCommonVendor = require('@plugin-light/webpack-plugin-add-common-vendor');

function _interopNamespace(e) {
    if (e && e.__esModule) return e;
    var n = Object.create(null);
    if (e) {
        Object.keys(e).forEach(function (k) {
            if (k !== 'default') {
                var d = Object.getOwnPropertyDescriptor(e, k);
                Object.defineProperty(n, k, d.get ? d : {
                    enumerable: true,
                    get: function () { return e[k]; }
                });
            }
        });
    }
    n["default"] = e;
    return Object.freeze(n);
}

var path__namespace = /*#__PURE__*/_interopNamespace(path);
var fs__namespace = /*#__PURE__*/_interopNamespace(fs);

function realTraverse(parent, deps, result, child) {
    for (const key of [...parent]) {
        if (result.indexOf(key) > -1) {
            continue;
        }
        result.push(key);
        let before = deps[key];
        if (before) {
            before = before.filter(ancestor => ancestor !== child && parent.indexOf(ancestor) <= -1);
            realTraverse(before, deps, result, child);
        }
    }
}
function excludeSubPackagePages({ deps, pages, mainPath, }) {
    return Object.keys(deps).reduce((acc, key) => {
        const isPage = pages.find(item => key.startsWith(item));
        if (isPage) {
            acc[key] = deps[key].filter(item => (item !== mainPath));
        }
        else {
            acc[key] = deps[key];
        }
        // acc[key] = deps[key].filter(item => (item !== mainPath || !isPage));
        return acc;
    }, {});
}
function excludeRepeatElements(deps) {
    return Object.keys(deps).reduce((acc, key) => {
        acc[key] = Array.from(new Set(deps[key])).filter(item => item !== key);
        return acc;
    }, {});
}
function traverseDeps({ deps, pages, mainPath, }) {
    deps = excludeRepeatElements(deps);
    deps = excludeSubPackagePages({ deps, pages, mainPath });
    const list = Object.keys(deps);
    const newDeps = {};
    for (const child of list) {
        const parent = [...deps[child]];
        const result = [];
        realTraverse([...parent], deps, result, child);
        newDeps[child] = result;
    }
    const res = excludeRepeatElements(newDeps);
    return res;
}

const getPageSet = () => sharedVue2.getUniCliCache('getPageSet')();
const INPUT_DIR = process.env.UNI_INPUT_DIR || '';
const SUB_PACKAGE_ROOTS = Object.keys(process.UNI_SUBPACKAGES || {});
const subPackageRootsWithSlash = Object.keys(process.UNI_SUBPACKAGES || {}).map(root => `${root}/`);
let mainPackagePages;
const ROOT_NAME = 'MAIN';
const VENDER_PATH = 'common/vendor';
/**
 * 获取主包页面，比如`src/project/user/views/index/index-home.vue`
 */
function getMainPackagePages() {
    if (mainPackagePages) {
        return mainPackagePages;
    }
    const subPackages = SUB_PACKAGE_ROOTS
        .map(item => path__namespace.resolve(INPUT_DIR, item));
    const pages = Array.from(getPageSet())
        .map(item => path__namespace.resolve(INPUT_DIR, item));
    mainPackagePages = pages
        .filter(page => !subPackages.find(subPackage => page.startsWith(subPackage)))
        .map(item => `${shared.getRelativePath(item)}.vue`);
    return mainPackagePages;
}
/**
 * 获取main.js路径，比如`src/project/user/main.js`
 */
function getMainJSPath() {
    return shared.getRelativePath(path__namespace.resolve(INPUT_DIR, 'main.js'));
}
/**
 * 获取页面列表
 */
function getPagesList() {
    const pages = Array.from(getPageSet());
    return pages.map(item => shared.getRelativePath(path__namespace.resolve(INPUT_DIR, item)));
}
/**
 * path是否为页面路径
 * @example
 * src/project/user/views/match-data/rank.vue // true
 * src/local-logic/tip-match/homepage/homepage.js // false
 */
function isPage(path) {
    const pages = getPagesList();
    const isPage = pages.find(item => path.startsWith(item));
    return !!isPage;
}
/**
 * 获取页面对应的分包
 */
function getPageSubPackages(page) {
    const subPackageMap = SUB_PACKAGE_ROOTS
        .reduce((acc, item) => {
        acc[shared.getRelativePath(path__namespace.resolve(INPUT_DIR, item))] = item;
        return acc;
    }, {});
    const subPackage = Object.keys(subPackageMap).find(item => page.startsWith(`${item}/`));
    if (subPackage) {
        return subPackageMap[subPackage];
    }
}
// 基础检测，vue和css类型不处理，只处理js类型
function baseTest(module, options) {
    const mainPath = tComm.normalizePath(path__namespace.resolve(process.env.UNI_INPUT_DIR || '', 'main.'));
    if (module.type === 'css/mini-extract') {
        return false;
    }
    if (module.resource) {
        const resource = tComm.normalizePath(module.resource);
        if (resource.indexOf('.vue') !== -1
            || resource.indexOf('.nvue') !== -1
            || resource.indexOf(mainPath) === 0 // main.js
        ) {
            return false;
        }
        if (options?.whiteList?.length) {
            const isExist = options.whiteList.find(url => resource.includes(url));
            return !isExist;
        }
    }
    return true;
}
const findSubPackages = function (chunks) {
    return chunks.reduce((pkgs, item) => {
        const name = tComm.normalizePath(item.name);
        // 用chunk的名称匹配分包
        const pkgRoot = subPackageRootsWithSlash.find(root => name.indexOf(root) === 0);
        pkgRoot && pkgs.add(pkgRoot);
        return pkgs;
    }, new Set());
};
const hasMainPackage = function (chunks) {
    // 有一个chunk的名称，在所有分包中都匹配不到
    return chunks.find(item => !subPackageRootsWithSlash.find(root => item.name.indexOf(root) === 0));
};
const findNameChunk = function (chunks, name) {
    return chunks.find(chunk => chunk.name === name);
};

const sourceDir = path__namespace.resolve(process.env.UNI_INPUT_DIR || '');
let scriptSubPackagesMap;
function getGlobalComponents() {
    const { globalCompsValues: globalComponents = [] } = webpackPluginDispatchVue.analyzeComponent({ needGlobalComponents: true });
    const sourceDirList = sourceDir.split(path__namespace.sep);
    return globalComponents
        // 过滤掉 wxcomponents 中的组件，因为它们没有依赖额外的 js
        .filter(item => item.indexOf('wxcomponents/') <= -1 && item.indexOf('plugin:/') <= -1)
        .map((item) => {
        if (item.indexOf('../..') <= -1)
            return item;
        const temp = item.replace('../../', '').replace(/node-modules/g, 'node_modules');
        for (let i = sourceDirList.length; i >= 0; i--) {
            const file = path__namespace.resolve(sourceDirList.slice(0, i).join('/'), `${temp}.vue`);
            if (fs__namespace.existsSync(file)) {
                return path__namespace.join(sourceDirList.slice(0, i).join('/'), `${temp}.vue`);
            }
        }
    })
        .map(item => shared.getRelativePath(item));
}
function isUsedInMainPackage(list) {
    const globalComponents = getGlobalComponents();
    const mainPackagePages = getMainPackagePages();
    const mainJSPath = getMainJSPath();
    shared.saveJsonToLog(mainPackagePages, 'dispatch-script.inner-main-package-pages.json');
    return list.find(item => [
        ROOT_NAME,
        mainJSPath,
        ...mainPackagePages,
        ...globalComponents,
    ].indexOf(item) > -1);
}
function getSubPackagesFromFiles(list) {
    let subPackages = list
        .filter(item => isPage(item))
        .map(item => getPageSubPackages(item))
        .filter(item => item);
    subPackages = Array.from(new Set(subPackages));
    return subPackages;
}
// 寻找 JS 对应的分包列表
function getScriptSubPackageMap(obj) {
    const keys = Object.keys(obj);
    // const res: Array<{js: Array<string>, subPackages: Array<string>}> = [];
    const parsedRes = {};
    const jsReg = /(.js|.ts)$/;
    for (const key of keys) {
        if (getPageSubPackages(key) || !jsReg.test(key)) {
            continue;
        }
        const list = obj[key];
        // const temp: Array<string> = [];
        if (isUsedInMainPackage(list)) {
            continue;
        }
        // temp.push(key);
        // temp.push(...list.filter((item: string) => jsReg.test(item) && !isPage(item)));
        const subPackages = getSubPackagesFromFiles(list);
        // res.push({
        //   js: temp,
        //   subPackages,
        // });
        parsedRes[key] = subPackages;
    }
    // saveJsonToLog(res, 'dispatch-script.inner-deps-to-dispatch.json');
    shared.saveJsonToLog(parsedRes, 'dispatch-script.inner-deps-to-dispatch-deep.json');
    return parsedRes;
}
/**
 * 获取一个 js/ts 文件对应的分包
 * @example
 * getScriptSubPackages('/Users/mike/Documents/web/src/project/user/mixins/tim-mixin.js', depsMap)
 * // [ 'views/sche/' ]
 */
function getScriptSubPackages(moduleName, depsMap) {
    if (!moduleName)
        return [];
    if (!scriptSubPackagesMap) {
        scriptSubPackagesMap = getScriptSubPackageMap(depsMap);
    }
    return scriptSubPackagesMap[shared.getRelativePath(moduleName)] || [];
}

/* eslint-disable @typescript-eslint/no-require-imports */
const resourceResolveDataList = [];
const moduleSources = [];
class DispatchScriptPlugin {
    options;
    moveFiles;
    pluginName = 'DispatchScriptPlugin';
    reverseDepsMap;
    handledModules;
    pages;
    mainPath;
    constructor(options) {
        this.options = options;
        this.moveFiles = new Map();
        this.pluginName = 'DispatchScriptPlugin';
        this.reverseDepsMap = {};
        this.handledModules = [];
        // 现在获取不到真正的 pages，是空数组
        this.pages = getPagesList();
        this.mainPath = getMainJSPath();
        shared.saveJsonToLog(this.pages, 'dispatch-script.raw-pages.json');
        shared.createLogDir();
    }
    apply(compiler) {
        compiler.hooks.normalModuleFactory.tap(this.pluginName, (nmf) => {
            nmf.hooks.afterResolve.tap(this.pluginName, (result) => {
                try {
                    this.collectDeps(result);
                }
                catch (err) {
                    console.warn('[DISPATCH SCRIPT] err', err);
                }
            });
        });
        compiler.hooks.thisCompilation.tap(this.pluginName, (compilation) => {
            compilation.hooks.optimizeChunksAdvanced.tap(this.pluginName, () => {
                try {
                    this.collectShouldDispatchJS(compilation);
                }
                catch (err) {
                    console.warn('[DISPATCH SCRIPT] err', err);
                }
            });
            compilation.hooks.optimizeChunkModules.tap(this.pluginName, (chunks, modules) => {
                shared.saveJsonToLog(resourceResolveDataList, 'dispatch-script.raw-resource-resolve-data-list.json');
                shared.saveJsonToLog(moduleSources, 'dispatch-script.raw-module-sources.json');
                try {
                    this.doDispatchJS({
                        modules,
                        chunks,
                        compilation,
                    });
                }
                catch (err) {
                    console.warn('[DISPATCH SCRIPT] err', err);
                }
            });
        });
        if (this.options.addCommonVendorRequire) {
            compiler.hooks.emit.tap(this.pluginName, (compilation) => {
                try {
                    const { assets } = compilation;
                    webpackPluginAddCommonVendor.addCommonVendorCore({
                        assets,
                        pageSet: Array.from(webpackPluginDispatchVue.getPageSet()),
                        subPackages: Object.keys(process.UNI_SUBPACKAGES) || {},
                        outputDir: process.env.UNI_OUTPUT_DIR || '',
                    });
                }
                catch (err) {
                    console.warn('DispatchScriptPlugin.err: ', err);
                }
            });
        }
    }
    collectShouldDispatchJS(compilation) {
        const reverseDeps = shared.parseSetDeps(this.reverseDepsMap);
        shared.saveJsonToLog(getPagesList(), 'dispatch-script.raw-pages-dispatch.json');
        const handledDepsMap = traverseDeps({
            deps: reverseDeps,
            pages: getPagesList(),
            mainPath: getMainJSPath(),
        });
        shared.saveJsonToLog(reverseDeps, 'dispatch-script.inner-deps-raw.json');
        shared.saveJsonToLog(handledDepsMap, 'dispatch-script.inner-deps-flatten.json');
        const waitDisposeModules = compilation.modules.filter((module) => baseTest(module, this.options));
        waitDisposeModules.forEach((module) => {
            const chunks = module.getChunks();
            const matchSubPackages = findSubPackages(chunks);
            const isMain = hasMainPackage(chunks);
            const forceMovePackages = getScriptSubPackages(module.resource, handledDepsMap);
            moduleSources.push(module.resource);
            if (forceMovePackages?.length) {
                this.moveFiles.set(module, {
                    name: module.resource,
                    pkgSet: forceMovePackages,
                });
            }
            else if (matchSubPackages.size > 0 && !isMain) {
                // 只用处理大于一个分包在使用的情况，一个使用的情况 uni 已经处理好了，但是要考虑 3 的情况
                this.moveFiles.set(module, {
                    name: module.resource,
                    pkgSet: matchSubPackages,
                });
            }
        });
    }
    doDispatchJS({ modules, chunks, compilation, }) {
        const graphHelpers = require('webpack/lib/GraphHelpers');
        modules.forEach((module) => {
            if (this.moveFiles.has(module)) {
                const mainChunks = module.getChunks();
                // 如果不存在主包的 common/vendor 中，说明已经被 uni 放入分包了，就不需要做任何处理
                if (mainChunks.length === 1 && mainChunks[0].name === VENDER_PATH) {
                    const mainChunk = mainChunks[0];
                    const moveFileInfo = this.moveFiles.get(module);
                    const chunkNames = [];
                    moveFileInfo.pkgSet.forEach((value) => {
                        /**
                         * pkgSet.value 举例：
                         *
                         * views/room/
                         * views/match/
                         * views/edit/
                         * views/setting/
                         */
                        const aPath = path__namespace.join(value, VENDER_PATH);
                        chunkNames.push(tComm.normalizePath(aPath));
                    });
                    chunkNames.forEach((chunkName) => {
                        let pkgChunk = findNameChunk(chunks, chunkName);
                        if (!pkgChunk) {
                            const group = compilation.addChunkInGroup(chunkName);
                            pkgChunk = group.chunks[0];
                        }
                        graphHelpers.connectChunkAndModule(pkgChunk, module);
                    });
                    this.handledModules.push({
                        module: shared.getRelativePath(module.resource),
                        subPackages: Array.from(moveFileInfo.pkgSet),
                    });
                    graphHelpers.disconnectChunkAndModule(mainChunk, module);
                }
            }
        });
        shared.saveJsonToLog(this.handledModules, 'dispatch-script.result-handled-modules.json');
    }
    collectDeps(result) {
        const { resourceResolveData } = result;
        resourceResolveDataList.push({
            ...resourceResolveData,
            descriptionFileData: {},
        });
        const { context: { issuer, compiler = '', }, path, query = '', } = resourceResolveData;
        const parent = issuer ? shared.getRelativePath(issuer) : ROOT_NAME;
        const child = shared.getRelativePath(path);
        if (parent === ROOT_NAME
            && typeof compiler === 'string'
            && compiler
            && compiler.indexOf('mini-css-extract-plugin') > -1
            && query
            && query.indexOf('&type=style') > -1) {
            return;
        }
        if (this.reverseDepsMap[child]) {
            this.reverseDepsMap[child].add(parent);
        }
        else {
            this.reverseDepsMap[child] = new Set([parent]);
        }
    }
}

exports.DispatchScriptPlugin = DispatchScriptPlugin;
